/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Zilibowitz on 2/01/05.
 */

#include "Player.h"
#include "GameConstants.h"

// constants

const double kShortToFrac = 1.0 / 65535.0;

const double kMaxSteerAngle = kPi / 4.0;			// rad
const double kSteerSpeed = kMaxSteerAngle*1.6;		// rad / s
const double kSteerReturnSpeed = kSteerSpeed*10;	// rad / s

const double kThrottleSpeed = 2.0;						// units / s
const double kThrottleReturnSpeed = kThrottleSpeed*10;	// units / s

const double kBrakeSpeed = 2.0;						// units / s
const double kBrakeReturnSpeed = kBrakeSpeed*10;	// units / s

// member functions

Player::Player(ToyCar& inCar)
	: car(inCar), lapFlag(-1), laps(0), curLap(0.0), prevLap(0.0), bestLapTime(INFINITY), courseTime(0.0)
{
	// init key states
	for (int i = 0; i < kNumKeys; i++)
		keys[i] = false;
}

void Player::doEvent(PlayerEvent* event)
{
	//assert(event != NULL);
	
	switch (event->type)
	{
		case kPlayerKeyUp:
			keyUp(((PlayerEventKey*)event)->key);
			break;
		
		case kPlayerKeyDown:
			keyDown(((PlayerEventKey*)event)->key);
			break;
		
		case kPlayerPaddleMove:
			paddleMove(((PlayerEventPaddle*)event)->control, ((PlayerEventPaddle*)event)->value);
			break;
	}
}

void Player::paddleMove(PlayerControl which, unsigned short value)
{
	switch (which)
	{
		case kSteerWheel:
			car.mSteerWheelAngle = (value*kShortToFrac - 0.5) * kMaxSteerAngle;
			break;
			
		case kThrottle:
			car.mThrottle = (value*kShortToFrac - 0.5) * 2.0;
			break;
			
		case kBrake:
			car.mBrake = value * kShortToFrac;
			break;
		
		default:
			fprintf(stderr, "Player::paddleMove received unknown control type %d\n", which);
	}
}

/* Updates the lap counter for this player. Returns 1 or 0 depending on whether
   the call to this function resulted in the lap count being increased or not.
*/
char Player::updateLaps(const Tuple& lineA, const Tuple& lineB, double curTime)
{
	const Tuple vec1 = car.getLocation_wc()*kScale - lineA;
	const Tuple vec2 = lineB - lineA;
	
	prevLap = curLap;
	curLap = vec2.Angle(vec1);
	
	if (curLap < 0)
		curLap += kTwoPi;	// move the branch cut thingy
	
	// check if we crossed over the branch cut in the atan function
	if (curLap - prevLap < -kPiOn2)
		lapFlag++;
	else if (curLap - prevLap > kPiOn2)
		lapFlag--;
	
	if (lapFlag == 1)
	{
		laps++;			// player has done a lap
		lapFlag = 0;
		
		// update time parameters
		bestLapTime = (curTime - courseTime) <? bestLapTime;
		courseTime = curTime;
		
		// test
		//printf("course time: %f\nbest lap time: %f\n", courseTime, bestLapTime);
        
        return 1;
	}
	else if (lapFlag == -2)
	{
		lapFlag = -1;	// player has done a lap backwards
	}
    
    return 0;
}

void Player::integrateKeys(double dt)
{
	// steering
	
	if (keys[kSteerWheelLeft])
	{
		car.mSteerWheelAngle += kSteerSpeed * dt;
		if (car.mSteerWheelAngle > kMaxSteerAngle)
			car.mSteerWheelAngle = kMaxSteerAngle;
	}
	
	if (keys[kSteerWheelRight])
	{
		car.mSteerWheelAngle -= kSteerSpeed * dt;
		if (car.mSteerWheelAngle < -kMaxSteerAngle)
			car.mSteerWheelAngle = -kMaxSteerAngle;
	}
	
	if (!(keys[kSteerWheelLeft] || keys[kSteerWheelRight]))
	{
		reduce_by_at_most(car.mSteerWheelAngle, kSteerReturnSpeed*dt);
	}
	
	// throttle

	if (keys[kThrottleMore])
	{
		car.mThrottle += kThrottleSpeed * dt;
		if (car.mThrottle > 1.0)
			car.mThrottle = 1.0;
	}
	
	if (keys[kThrottleLess])
	{
		car.mThrottle -= kThrottleSpeed * dt;
		if (car.mThrottle < -1.0)
			car.mThrottle = -1.0;
	}
	
	if (!(keys[kThrottleMore] || keys[kThrottleLess]))
	{
		reduce_by_at_most(car.mThrottle, kThrottleReturnSpeed*dt);
	}
	
	// brake

	if (keys[kBrakeMore])
	{
		car.mBrake += kBrakeSpeed * dt;
		if (car.mBrake > 1.0)
			car.mBrake = 1.0;
	}
	
	if (keys[kBrakeLess])
	{
		car.mBrake -= kBrakeSpeed * dt;
		if (car.mBrake < 0.0)
			car.mBrake = 0.0;
	}
	
	if (!(keys[kBrakeMore] || keys[kBrakeLess]))
	{
		reduce_by_at_most(car.mBrake, kBrakeReturnSpeed*dt);
	}
}
